#!/usr/bin/env python3
# THIS "versionfile" IS GENERATED BY version-pioneer-0.0.8
# by evaluating the original versionscript and storing the computed versions as a constant.

def get_version_dict():
    return {'version': '0.0.8', 'full_revisionid': 'a68da113234a5eb656308871480254f343986aa0', 'dirty': False, 'error': None, 'date': '2024-12-26T15:57:26+0000'}


if __name__ == "__main__":
    import json

    print(json.dumps(get_version_dict()))